﻿using HarmonyLib;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Runtime.Serialization;
using System.Reflection;
using UnityEngine;
using FMODUnity;
using Steamworks;

public class SimpleCharger : Mod
{
    public static AssetBundle asset;
    private Harmony harmony;

    public void Start()
    {
        try
        {
            HNotification notification = FindObjectOfType<HNotify>()
                .AddNotification(HNotify.NotificationType.spinning, "Loading Simple Charger...");

            var bundleLoadRequest = AssetBundle.LoadFromMemoryAsync(
                GetEmbeddedFileBytes("simple_charger.assets")
            );
            asset = bundleLoadRequest.assetBundle;

            var Simple_Charger = asset.LoadAsset<Item_Base>("Placeable_Simple_Charger");

            RAPI.AddItemToBlockQuadType(Simple_Charger, RBlockQuadType.quad_foundation);
            RAPI.AddItemToBlockQuadType(Simple_Charger, RBlockQuadType.quad_foundation_empty);
            RAPI.RegisterItem(Simple_Charger);

            foreach (Block b in Simple_Charger.settings_buildable.GetBlockPrefabs())
            {
                b.networkedBehaviour = b.gameObject.AddComponent<SimpleCharger_Network>();
                b.networkType = NetworkType.NetworkBehaviour;
            }

            Simple_Charger.SetRecipe(
                new[]
                {
                    new CostMultiple(new[] { ItemManager.GetItemByIndex(21) }, 4),
                    new CostMultiple(new[] { ItemManager.GetItemByIndex(22) }, 4),
                    new CostMultiple(new[] { ItemManager.GetItemByIndex(23) }, 4),
                    new CostMultiple(new[] { ItemManager.GetItemByIndex(178) }, 2)
                },
                CraftingCategory.Other,
                1,
                false,
                "BatteryCharger",
                10
            );

            this.harmony = new Harmony("com.plateglassarmour.Simple-Charger");
            this.harmony.PatchAll(Assembly.GetExecutingAssembly());

            notification.Close();
            Debug.Log("Simple Charger has been successfully loaded!");
        }
        catch (Exception e)
        {
            Debug.LogError(e);
        }
    }

    public void OnModUnload()
    {
        asset.Unload(true);
        this.harmony?.UnpatchAll("com.plateglassarmour.Simple-Charger");
        Debug.Log("Simple Charger has been unloaded!");
    }
}

static class ExtentionMethods
{
    public static void SetRecipe(
        this Item_Base item,
        CostMultiple[] cost,
        CraftingCategory category = CraftingCategory.Resources,
        int amountToCraft = 1,
        bool learnedFromBeginning = false,
        string subCategory = null,
        int subCatergoryOrder = 0
    )
    {
        Traverse recipe = Traverse.Create(item.settings_recipe);
        recipe.Field("craftingCategory").SetValue(category);
        recipe.Field("amountToCraft").SetValue(amountToCraft);
        recipe.Field("learnedFromBeginning").SetValue(learnedFromBeginning);
        recipe.Field("subCategory").SetValue(subCategory);
        recipe.Field("subCatergoryOrder").SetValue(subCatergoryOrder);
        item.settings_recipe.NewCost = cost;
    }
}

[HarmonyPatch(typeof(RGD_Block), "RestoreBlock")]
public class Patch_SimpleChargerRestoreBlock
{
    public static void Postfix(ref RGD_Block __instance, Block block)
    {
        if (__instance.BlockIndex == 16961)
        {
            RGD_Storage rgd = __instance as RGD_Storage;
            SimpleCharger_Network dn = block.networkedBehaviour as SimpleCharger_Network;
            if (rgd != null && dn != null)
            {
                dn.RestoreData(rgd);
            }
        }
    }
}

[HarmonyPatch(typeof(Block), "GetBlockCreationData")]
public class Patch_BlockCreateSaveData
{
    public static bool Prefix(ref Block __instance, ref RGD_Block __result)
    {
        if (__instance.GetComponent<SimpleCharger_Network>() != null)
        {
            __result =
                __instance.GetComponent<SimpleCharger_Network>().Serialize_Save() as RGD_Block;
            return false;
        }
        return true;
    }
}

public static class ExtensionMethods
{
    public static void CopyFieldsOf(this object value, object source)
    {
        var t1 = value.GetType();
        var t2 = source.GetType();
        while (!t1.IsAssignableFrom(t2))
            t1 = t1.BaseType;
        while (t1 != typeof(UnityEngine.Object) && t1 != typeof(object))
        {
            foreach (var f in t1.GetFields(~BindingFlags.Default))
                if (!f.IsStatic)
                    f.SetValue(value, f.GetValue(source));
            t1 = t1.BaseType;
        }
    }
}

class SimpleCharger_Network : MonoBehaviour_Network
{
    public Battery battery;

    protected Animator animator;
    protected Raft_Network network;
    protected bool isPlaced = false;
    protected Raft raft;
    protected float timer = 0;
    protected int batteryChargesPerTick = 1;

    public virtual void OnBlockPlaced()
    {
        NetworkIDManager.AddNetworkID(this);
        this.network = ComponentManager<Raft_Network>.Value;
        this.raft = ComponentManager<Raft>.Value;
        this.isPlaced = true;
        Battery battery = this.GetComponentInChildren<Battery>();
        this.animator = this.GetComponentInChildren<Animator>();
        this.battery = battery;
        Traverse.Create(this.battery).Field("networkBehaviourID").SetValue(this);
    }

    protected override void OnDestroy()
    {
        base.OnDestroy();
        NetworkIDManager.RemoveNetworkID(this);
    }

    private void Update()
    {
        if (!this.isPlaced)
        {
            return;
        }
        this.HandleAnimation();
        if (Raft_Network.IsHost)
        {
            if (
                battery != null
                && !battery.BatterySlotIsEmpty
                && battery.NormalizedBatteryLeft != 1f
                && !raft.IsAnchored
            )
            {
                this.timer += Time.deltaTime;
                if (this.timer >= 6f)
                {
                    this.timer -= 6f;
                    this.battery.AddBatteryUsesNetworked(this.batteryChargesPerTick);
                }
            }
        }
    }

    private void HandleAnimation()
    {
        if (this.animator == null)
        {
            return;
        }
        if (!this.raft.IsAnchored)
        {
            this.animator.SetFloat("WheelSpeed", 1f);
            this.animator.SetFloat("BarrelSpeed", 1f);
        }
        else
        {
            this.animator.SetFloat("WheelSpeed", 0f);
            this.animator.SetFloat("BarrelSpeed", 0f);
        }
    }

    public override bool Deserialize(Message_NetworkBehaviour msg, CSteamID remoteID)
    {
        switch (msg.Type)
        {
            case Messages.Battery_Insert:
            case Messages.Battery_Take:
            case Messages.Battery_Update:
            case Messages.Battery_OnOff:
                return this.battery.OnBatteryMessage(msg);
            default:
                return base.Deserialize(msg, remoteID);
        }
    }

    public void RestoreData(RGD_Storage rgd)
    {
        if (rgd.slots[0].itemUses == 1)
        {
            byte[] data = Convert.FromBase64String(rgd.slots[0].exclusiveString);
            int uses = BitConverter.ToInt32(data, 0);
            int index = BitConverter.ToInt32(data, 4);
            this.battery.Insert(null, uses, index);
        }
    }

    public override RGD Serialize_Save()
    {
        RGD_Storage rgd = CreateObject<RGD_Storage>();
        rgd.CopyFieldsOf(new RGD_Block(RGDType.Block, this.GetComponent<Block>()));
        rgd.slots = new RGD_Slot[] { CreateObject<RGD_Slot>() };

        if (this.battery == null || this.battery.GetBatteryInstance() == null)
        {
            rgd.slots[0].itemUses = 0;
        }
        else
        {
            rgd.slots[0].itemUses = 1;
            List<byte> data = new List<byte>();
            data.AddRange(BitConverter.GetBytes(this.battery.BatteryUses));
            data.AddRange(BitConverter.GetBytes(this.battery.GetBatteryInstance().UniqueIndex));

            rgd.slots[0].exclusiveString = Convert.ToBase64String(data.ToArray());
        }
        return rgd;
    }

    public static T CreateObject<T>() => (T)FormatterServices.GetUninitializedObject(typeof(T));
}
